/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/cpp_gen/cpp_proxy_body_gen.h"

namespace {
#include "cpp_proxy_body_gen_cb.h"
}

namespace tidl {

CppProxyBodyGen::CppProxyBodyGen(std::shared_ptr<Document> doc)
    : CppGeneratorBase(doc) {}

void CppProxyBodyGen::OnInitGen(std::ofstream& stream) {
  std::string key(".cc");
  std::string header_file = FileName;

  std::size_t found = header_file.rfind(key);
  if (found != std::string::npos)
    header_file.replace(found, key.length(), ".h");

  GenCopyright(stream);
  stream << NLine(1);
  stream << "#include <stdlib.h>" << NLine(1)
         << "#include <assert.h>" << NLine(1)
         <<  NLine(1)
         << "#include \"" << header_file << "\"" << NLine(2);
  GenNamespace(stream);
}

void CppProxyBodyGen::OnFiniGen(std::ofstream& stream) {
}

void CppProxyBodyGen::GenNamespace(std::ofstream& stream) {
  stream << "namespace rpc_port ";
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << " ";
    GenBrace(stream, 0, [&]() {
      stream <<  NLine(1);
      GenStructuresForBody(stream);
      stream << "namespace proxy ";
      GenBrace(stream, 0, [&]() {
        GenPrototype(stream);
        GenSerializer(stream);
        GenDeSerializer(stream);
        GenListSerializer(stream);
        GenInterfaces(stream);
      }, false);
    }, false);
  }, false);
}

void CppProxyBodyGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;
    Interface& iface = static_cast<Interface&>(*i);
    GenInterface(stream, iface);
  }
}

void CppProxyBodyGen::GenInterface(std::ofstream& stream,
                                   const Interface& iface) {
  GenBodyCallbacks(stream, iface, true);
  GenConstructor(stream, iface);
  GenDestructor(stream, iface);
  GenHelperMethods(stream, iface);
  GenMethods(stream, iface);
}

void CppProxyBodyGen::GenConstructor(std::ofstream& stream,
                                     const Interface& iface) {
  GenTemplate(CB_PROXY_INTERFACE_CTOR, stream,
    [&]()->std::string {
      return iface.GetID();
    },
    [&]()->std::string {
      return iface.GetID();
    }
  );
}

void CppProxyBodyGen::GenDestructor(std::ofstream& stream,
                                    const Interface& iface) {
  GenTemplate(CB_DTOR, stream,
    [&]()->std::string {
      return iface.GetID();
    },
    [&]()->std::string {
      return iface.GetID();
    }
  );
}

void CppProxyBodyGen::GenHelperMethods(std::ofstream& stream,
                                       const Interface& iface) {
  stream << ReplaceAll(CB_PROXY_HELPER_METHODS, "##", iface.GetID())
         << NLine(1);
}

void CppProxyBodyGen::GenMethods(std::ofstream& stream,
                                 const Interface& iface) {
  auto& decls = iface.GetDeclarations();

  for (auto& i : decls.GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;

    GenDeclaration(stream, iface, *i);
    GenBrace(stream, 0, [&]() {
      GenInvocation(stream, *i);
    }, false);
    stream << NLine(1);
  }
}

void CppProxyBodyGen::GenInvocation(std::ofstream& stream, const Declaration& decl) {
  stream << CB_INVOCATION_PRE;

  // Serialize
  stream << Tab(1)
         << "rpc_port_parcel_write_int32(p, static_cast<int>(MethodId::"
         << decl.GetID() << "));" << NLine(1);
  std::string m;
  std::string l;
  for (auto& i : decl.GetParameters().GetParams()) {
    auto& pt = i->GetParameterType();
    if (pt.GetDirection() == ParameterType::Direction::OUT)
      continue;
    m += ConvertTypeToSerializer(pt.GetBaseType(), i->GetID(), "p");
    if (IsDelegateType(pt.GetBaseType())) {
      l += "delegate_list_.emplace_back(" + i->GetID() + ".release());\n";
    }
  }
  stream << AddIndent(TAB_SIZE, m) << NLine(1);

  if (decl.GetMethodType() == Declaration::MethodType::SYNC)
    stream << Tab(1) << "rpc_port_parcel_h parcel_received;" << NLine(1);
  stream << Tab(1) << "do ";
  GenBrace(stream, TAB_SIZE, [&]() {
    stream << Tab(2) << "std::lock_guard<std::recursive_mutex> lock(mutex_);"
           << NLine(2);
    if (!l.empty())
      stream << AddIndent(TAB_SIZE * 2, l);
    stream << CB_INVOCATION_MID << NLine(1);
    if (decl.GetMethodType() == Declaration::MethodType::SYNC) {
      stream << Tab(2) << "// Receive" << NLine(1);
      stream << Tab(2)
             << "ConsumeCommand(&parcel_received, port_);" << NLine(1);
    }
  }, false, false);
  stream << " while (false);" << NLine(1);

  // Deserialize
  if (decl.GetMethodType() == Declaration::MethodType::ASYNC) {
    stream << Tab(1) << "rpc_port_parcel_destroy(p);"
         << NLine(1);
    return;
  }

  stream << CB_INVOCATION_RECEIVE << NLine(1);
  for (auto& i : decl.GetParameters().GetParams()) {
    if (i->GetParameterType().GetDirection() == ParameterType::Direction::IN) {
      continue;
    }

    std::string c = ConvertTypeToDeserializer(
        i->GetParameterType().GetBaseType(),
        i->GetID(), "parcel_received", false);
    if (c != "")
      stream << AddIndent(TAB_SIZE, c);
  }

  if (decl.GetType().ToString() != "void") {
    stream << AddIndent(TAB_SIZE,
                        ConvertTypeToDeserializer(decl.GetType(),
                                                "ret", "parcel_received"));
  }

  stream << CB_INVOCATION_END;
}

}  // namespace tidl
