/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/cpp_gen/cpp_proxy_header_gen.h"

namespace {
#include "cpp_proxy_header_gen_cb.h"
}

namespace tidl {

CppProxyHeaderGen::CppProxyHeaderGen(std::shared_ptr<Document> doc)
    : CppGeneratorBase(doc) {}

void CppProxyHeaderGen::OnInitGen(std::ofstream& stream) {
  GenCopyright(stream);
  stream << CB_HEADER;
  GenNamespace(stream);
}

void CppProxyHeaderGen::OnFiniGen(std::ofstream& stream) {
}

void CppProxyHeaderGen::GenNamespace(std::ofstream& stream) {
  stream << "namespace rpc_port ";
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << " ";
    GenBrace(stream, 0, [&]() {
      stream <<  NLine(1);
      GenStructuresForHeader(stream);
      stream << "namespace proxy ";
      GenBrace(stream, 0, [&]() {
        GenExceptions(stream);
        GenInterfaces(stream);
      }, false);
    }, false);
  }, false);
}

void CppProxyHeaderGen::GenExceptions(std::ofstream& stream) {
  stream << CB_EXCEPTIONS;
}

void CppProxyHeaderGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;
    Interface& iface = static_cast<Interface&>(*i);
    GenInterface(stream, iface);
  }
}

void CppProxyHeaderGen::GenInterface(std::ofstream& stream,
                                     const Interface& iface) {
  stream << NLine(1) << "class " << iface.GetID() << " ";
  GenBrace(stream, 0, [&]() {
    stream << " public:" << NLine(1);
    GenHeaderCallbacks(stream, iface, true);
    GenTemplate(CB_PUBLIC_MEMBERS, stream,
      [&]()->std::string {
        return iface.GetID();
      },
      [&]()->std::string {
        return iface.GetID();
      }
    );
    GenMethods(stream, iface);
    stream << NLine(1) << " private:" << NLine(1);
    GenMethodId(stream, iface);
    GenDelegateId(stream, iface);
    stream << CB_PRIVATE_MEMBERS;
  }, false, false);
  stream << ";" << NLine(1);
}

void CppProxyHeaderGen::GenMethods(std::ofstream& stream,
                                   const Interface& iface) {
  auto& decls = iface.GetDeclarations();

  for (auto& i : decls.GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;

    GenDeclaration(stream, *i);
  }
}

void CppProxyHeaderGen::GenDeclaration(std::ofstream& stream,
                                       const Declaration& decl) {
  if (!decl.GetComments().empty())
    stream << NLine(1) << AddIndent(TAB_SIZE, decl.GetComments());

  stream << Tab(1) << ConvertTypeToString(decl.GetType()) << " "
         << decl.GetID() << "(";
  GenParameters(stream, decl.GetParameters());
  stream << ");" << NLine(1);
}
}  // namespace tidl
