/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

const char CB_EXCEPTIONS[] =
R"__cpp_cb(
class Exception {};
class NotConnectedSocketException : public Exception {};
class InvalidProtocolException : public Exception {};
class InvalidIOException : public Exception {};
class PermissionDeniedException : public Exception {};
class InvalidIDException : public Exception {};
)__cpp_cb";

const char CB_PUBLIC_MEMBERS[] =
R"__cpp_cb(
  class IEventListener {
   public:
    /// <summary>
    /// This method will be invoked when the client app is connected to the servicece app.
    /// </summary>
    virtual void OnConnected() = 0;

    /// <summary>
    /// This method will be invoked after the client app was disconnected from the servicece app.
    /// </summary>
    virtual void OnDisconnected() = 0;

    /// <summary>
    /// This method will be invoked when the service app rejects the client app.
    /// </summary>
    virtual void OnRejected() = 0;
  };

  /// <summary>
  /// Constructor for this class
  /// </summary>
  /// <param name="listener">The listener for events</param>
  /// <param name="target_appid">The service app ID to connect</param>
  $$(IEventListener* listener, const std::string& target_appid);

  /// <summary>
  /// Destructor for this class
  /// </summary>
  virtual ~$$();

  /// <summary>
  /// Connects to the service app.
  /// </summary>
  /// <privilege>http://tizen.org/privilege/appmanager.launch</privilege>
  /// <privilege>http://tizen.org/privilege/datasharing</privilege>
  /// <exception cref="InvalidIDException">
  /// Thrown when the appid to connect is invalid.
  /// </exception>
  /// <exception cref="InvalidIOException">
  /// Thrown when internal I/O error happen.
  /// </exception>
  /// <exception cref="PermissionDeniedException">
  /// Thrown when the permission is denied.
  /// </exception>
  /// <remark> If you want to use this method, you must add privileges.</remark>
  void Connect();

  /// <summary>
  /// Disposes delegate objects in this interface
  /// </summary>
  /// <param name="tag">The tag string from delegate object</param>
  void DisposeCallback(const std::string& tag);

)__cpp_cb";

const char CB_PRIVATE_MEMBERS[] =
R"__cpp_cb(  void ProcessReceivedEvent(rpc_port_parcel_h parcel);
  void ConsumeCommand(rpc_port_parcel_h* parcel, rpc_port_h port);

  static void OnConnectedCB(const char *ep, const char *port_name,
                            rpc_port_h port, void *data);
  static void OnDisconnectedCB(const char *ep, const char *port_name,
                               void *data);
  static void OnRejectedCB(const char *ep, const char *port_name, void *data);
  static void OnReceivedCB(const char *ep, const char *port_name, void *data);

  rpc_port_h port_;
  rpc_port_proxy_h proxy_;
  IEventListener* listener_;
  std::recursive_mutex mutex_;
  std::list<std::unique_ptr<CallbackBase>> delegate_list_;
  std::string target_appid_;
)__cpp_cb";

const char CB_HEADER[] =
R"__cpp_cb(
#pragma once

#include <bundle.h>
#include <rpc-port-parcel.h>
#include <rpc-port.h>

#include <string>
#include <vector>
#include <memory>
#include <mutex>
#include <list>

)__cpp_cb";
