/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "idlc/cs_gen/cs_proxy_gen.h"

namespace {
#include "idlc/cs_gen/cs_cb_interop.h"
#include "idlc/cs_gen/cs_cb_proxy_interop.h"
#include "idlc/cs_gen/cs_proxy_gen_cb.h"
}

namespace tidl {

CsProxyGen::CsProxyGen(std::shared_ptr<Document> doc)
    : CsGeneratorBase(doc) {}

void CsProxyGen::OnInitGen(std::ofstream& stream) {
  GenCopyright(stream);
  stream << "using System;" << NLine(1)
         << "using System.Collections.Generic;" << NLine(1)
         << "using System.Runtime.InteropServices;" << NLine(1)
         << "using Tizen.Applications;" << NLine(1);
  GenInterop(stream);
  GenNamespace(stream);
}

void CsProxyGen::OnFiniGen(std::ofstream& stream) {
}

void CsProxyGen::GenNamespace(std::ofstream& stream) {
  stream << "namespace RPCPort" << NLine(1);
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << NLine(1);
    GenBrace(stream, 0, [&]() {
      GenExceptions(stream);
      GenStructures(stream);
      stream << Tab(1) << "namespace Proxy" << NLine(1);
      GenBrace(stream, TAB_SIZE, [&]() {
        GenInterfaces(stream);
      });
    });
  });
}

void CsProxyGen::GenInterop(std::ofstream& stream) {
  stream << "namespace RPCPort" << NLine(1);
  GenBrace(stream, 0, [&]() {
    stream << "namespace " << GetFileNamespace() << NLine(1);
    GenBrace(stream, 0, [&]() {
      GenTemplate(AddIndent(TAB_SIZE, ::cs_cb_interop), stream,
        [&]()->std::string {
          return AddIndent(TAB_SIZE, ::cs_cb_proxy_interop);
        }
      );
    });
  });
  stream << NLine(1);
}

void CsProxyGen::GenInterfaces(std::ofstream& stream) {
  for (auto& i : GetDocument().GetBlocks()) {
    if (i->GetType() != Block::TYPE_INTERFACE)
      continue;
    Interface& iface = static_cast<Interface&>(*i);
    GenInterface(stream, iface);
    stream << NLine(1);
  }
}

void CsProxyGen::GenInterface(std::ofstream& stream, const Interface& iface) {

  stream << Tab(2) << "public class " << iface.GetID()
         << " : IDisposable" << NLine(1);
  GenBrace(stream, TAB_SIZE * 2, [&]() {
    stream << CB_DATA_MEMBERS;
    GenCallbacks(stream, iface, true);
    GenDelegateId(stream, iface);
    GenMethodId(stream, iface);
    stream << CB_EVENT_METHODS;
    GenSerializer(stream);
    GenListSerializer(stream);
    GenCtor(stream, iface);
    GenConnectMethod(stream, iface);
    GenMethods(stream, iface);
    GenDisposable(stream, iface);
  });
}

void CsProxyGen::GenCtor(std::ofstream& stream, const Interface& iface) {
  const char* m = "public $$(string appId) => _appId = appId;\n";

  stream << NLine(1);
  GenTemplate(AddIndent(TAB_SIZE * 3, m), stream,
    [&]()->std::string {
      return iface.GetID();
    }
  );
}

void CsProxyGen::GenConnectMethod(std::ofstream& stream, const Interface& iface) {
  GenTemplate(CB_CONNECT_METHOD, stream,
    [&]()->std::string {
      return iface.GetID();
    }
  );
  stream << NLine(1);
}

void CsProxyGen::GenMethods(std::ofstream& stream, const Interface& iface) {
  auto& decls = iface.GetDeclarations();

  for (auto& i : decls.GetDecls()) {
    if (i->GetMethodType() == Declaration::MethodType::DELEGATE)
      continue;

    if (!i->GetComments().empty())
      stream << AddIndent(TAB_SIZE * 3, i->GetComments());

    stream << Tab(3) << "public ";
    GenDeclaration(stream, *i, false);
    stream << NLine(1);
    GenBrace(stream, TAB_SIZE * 3, [&]() {
      GenInvocation(stream, *i);
    });
    stream << NLine(1);
  }
}

void CsProxyGen::GenInvocation(std::ofstream& stream, const Declaration& decl) {
  stream << CB_INVOCATION_PRE;

  // Serialize
  stream << Tab(4)
         << "Interop.LibRPCPort.Parcel.WriteInt32(p, (int)MethodId."
         << decl.GetID() << ");" << NLine(1);
  std::string m;
  std::string l;
  for (auto& i : decl.GetParameters().GetParams()) {
    auto& pt = i->GetParameterType();
    if (pt.GetDirection() == ParameterType::Direction::OUT)
      continue;
    m += ConvertTypeToSerializer(pt.GetBaseType(), i->GetID(), "p");
    if (IsDelegateType(pt.GetBaseType())) {
      l += "_delegateList.Add(" + i->GetID() + ");\n";
    }
  }
  stream << AddIndent(TAB_SIZE * 4, m) << NLine(1);

  if (decl.GetMethodType() == Declaration::MethodType::SYNC)
    stream << Tab(4) << "IntPtr parcel_received;" << NLine(1);
  stream << Tab(4) << "lock (_lock)" << NLine(1);
  GenBrace(stream, TAB_SIZE * 4, [&]() {
    if (!l.empty())
      stream << AddIndent(TAB_SIZE * 5, l) << NLine(1);
    stream << CB_INVOCATION_MID << NLine(1);
    if (decl.GetMethodType() == Declaration::MethodType::SYNC) {
      stream << Tab(5) << "// Receive" << NLine(1);
      stream << Tab(5)
             << "ConsumeCommand(out parcel_received, _port);" << NLine(1);
    }
  });
  stream << NLine(1);

  // Deserialize
  if (decl.GetMethodType() == Declaration::MethodType::ASYNC) {
    stream << Tab(4) << "Interop.LibRPCPort.Parcel.Destroy(p);"
         << NLine(1);
    return;
  }

  const char* receive_block =
      "if (parcel_received == IntPtr.Zero)\n" \
      "{\n" \
      "    throw new InvalidProtocolException();\n" \
      "}\n";
  stream << AddIndent(TAB_SIZE * 4, receive_block) << NLine(1);

  for (auto& i : decl.GetParameters().GetParams()) {
    if (i->GetParameterType().GetDirection() == ParameterType::Direction::IN) {
      continue;
    }

    std::string c = ConvertTypeToDeserializer(
        i->GetParameterType().GetBaseType(),
        i->GetID(), "parcel_received", false);
    if (c != "")
      stream << AddIndent(TAB_SIZE * 4, c);
  }

  if (decl.GetType().ToString() != "void") {
    stream << AddIndent(TAB_SIZE * 4,
                        ConvertTypeToDeserializer(decl.GetType(),
                                                "ret", "parcel_received"));
  }

  stream << Tab(4) << "Interop.LibRPCPort.Parcel.Destroy(p);"
         << NLine(1);
  stream << Tab(4) << "Interop.LibRPCPort.Parcel.Destroy(parcel_received);"
         << NLine(1);
  stream << Tab(4) << "return ret;" << NLine(1);
}

void CsProxyGen::GenDisposable(std::ofstream& stream, const Interface& iface) {
  GenTemplate(CB_DISPOSABLE, stream,
    [&]()->std::string {
      return iface.GetID();
    }
  );
}

}  // namespace tidl
