/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <fstream>
#include <string>

#include "idlc/parser.h"
#include "idlc/document.h"
#include "idlc/declaration.h"
#include "idlc/parameter.h"
#include "idlc/type.h"
#include "idlc/element.h"
#include "idlc/structure.h"
#include "idlc/tidlc_y.hpp"

struct yy_buffer_state;
yy_buffer_state* yy_scan_buffer(char*, size_t, void*);
void yylex_init(void**);
void yylex_destroy(void*);

namespace tidl {

Parser::Parser() : scanner_(nullptr) {
  yylex_init(&scanner_);
}

Parser::~Parser() {
  yylex_destroy(scanner_);
}

void Parser::SetDoc(Document* doc) {
  doc_ = std::shared_ptr<Document>(doc);
}

std::shared_ptr<Document> Parser::GetDoc() {
  return doc_;
}

bool Parser::Parse(const std::string& input) {
  std::string in = input;
  error_ = false;

  in.append(2u, '\0');
  yy_scan_buffer(&in[0], in.length(), scanner_);
  if (yy::parser(this).parse() != 0 || error_ != false) {
    std::cerr << "[TIDL:Parser] error" << std::endl;
    return false;
  }

  return true;
}

bool Parser::ParseFromFile(const std::string& path) {
  std::ifstream inFile(path);
  std::string str((std::istreambuf_iterator<char>(inFile)),
      std::istreambuf_iterator<char>());
  path_ = path;
  return Parse(str);
}

void Parser::ReportError(const std::string& err, unsigned line) {
  std::cerr << path_ << ":" << line << ": " << err << std::endl;
  error_ = true;
}

}  // namespace tidl
